from verification_digit_pty.plugins import _PLUGINS


def test_plugin_register_plugin() -> None:
    from verification_digit_pty.plugins import register_plugin

    @register_plugin
    def f() -> str:
        return "f"

    assert f() == "f"
    assert "tests.unit.test_plugins.f" in _PLUGINS



from pathlib import Path

import pytest


@pytest.fixture(scope="session")
def output_folder() -> Path:
    return Path(__file__).parent.parent / "output"


@pytest.fixture(scope="session")
def fixtures_folder() -> Path:
    return Path(__file__).parent / "fixtures"



import inspect
from typing import Any, Callable

_PLUGINS = {}


def get_plugin(plugin: str) -> Callable[..., Any]:
    plugins_found = []
    for key in _PLUGINS:
        if plugin in key:
            plugins_found.append(_PLUGINS[key])
    if len(plugins_found) == 1:
        return plugins_found[0]
    elif len(plugins_found) > 1:
        raise ValueError(f"Plugin {plugin} is ambiguous")
    else:
        raise ValueError(f"Plugin {plugin} not found")


def register_plugin(func: Callable[..., Any]) -> Callable[..., Any]:
    module = get_module_name(func)
    plugin_key = f"{module}.{func.__name__}"
    if plugin_key in _PLUGINS:
        raise ValueError(f"Plugin {plugin_key} already registered")
    _PLUGINS[plugin_key] = func
    return func


def get_module_name(func):
    return inspect.getmodule(func).__name__



from verification_digit_pty.enums import NaturalRUCLetter, Province
from verification_digit_pty.exceptions import VerificationDigitError
from verification_digit_pty.plugins import register_plugin


def e_adapter_old(ruc: str) -> str:
    ruc_parts = ruc.split("-")
    try:
        ructb = (
            "0" * (4 - len(ruc_parts[1]))
            + "0000005"
            + "00"
            + "50"
            + "0" * (3 - len(ruc_parts[1]))
            + ruc_parts[1]
            + "0" * (5 - len(ruc_parts[2]))
            + ruc_parts[2]
        )
    except IndexError:
        raise VerificationDigitError("Invalid RUC") from None
    return ructb


@register_plugin
def e_adapter(ruc: str) -> str:
    ruc_parts = ruc.split("-")
    letter = NaturalRUCLetter.from_code(ruc_parts[0])
    folio_imagen = ruc_parts[1]
    asiento_ficha = ruc_parts[2]
    max_folio_len = 4

    if 0 < len(folio_imagen) < max_folio_len:
        if len(asiento_ficha) == 6 and letter in (NaturalRUCLetter.E, NaturalRUCLetter.N):
            ructb = ("5" + "00" + letter.code.ljust(2, "0") + folio_imagen.zfill(3) + asiento_ficha).zfill(20)
        else:
            ructb = ("5" + "00" + letter.code.ljust(2, "0") + folio_imagen.zfill(3) + asiento_ficha[:5].zfill(5)).zfill(
                20
            )
    elif len(asiento_ficha) == 6 and letter in (NaturalRUCLetter.E, NaturalRUCLetter.N):
        ructb = (
            "5" + letter.validation_code + letter.code.ljust(2, "0") + folio_imagen.zfill(4) + asiento_ficha
        ).zfill(20)
    else:
        ructb = (
            "5" + letter.validation_code + letter.code + folio_imagen.zfill(4) + asiento_ficha[:5].zfill(5)
        ).zfill(20)
    return ructb


def n_adapter(ruc: str) -> str:
    return e_adapter(ruc)


def av_adapter(ruc: str) -> str:
    ruc_parts = ruc.split("-")
    letter = NaturalRUCLetter.from_part(ruc_parts[0])
    folio_imagen = ruc_parts[1]
    asiento_ficha = ruc_parts[2]
    max_folio_len = 4
    province = Province.from_part(ruc_parts[0])
    if 0 < len(folio_imagen) < max_folio_len:
        ructb = ("5" + province.code.zfill(2) + letter.code + folio_imagen.zfill(3) + asiento_ficha[:5].zfill(5)).zfill(
            20
        )
    else:
        ructb = ("5" + letter.validation_code + folio_imagen.zfill(4) + asiento_ficha[:5].zfill(5)).zfill(20)
    return ructb


def pi_adapter(ruc: str) -> str:
    return av_adapter(ruc)


def pe_adapter(ruc: str) -> str:
    return e_adapter(ruc)


def nt_adapter(ruc: str) -> str:
    ruc = ruc.replace(" ", "").upper()

    if ruc.count("-") != 3:
        raise ValueError("Formato de RUC Natural NT incorrecto.")

    ruc_parts = ruc.split("-")

    if len(ruc_parts) != 4 or ruc_parts[1] != "NT":
        raise ValueError("Formato de RUC Natural NT incorrecto.")

    province = Province.from_code(ruc_parts[0])
    folio_imagen = ruc_parts[2]
    asiento_ficha = ruc_parts[3]
    if len(asiento_ficha) <= 5:
        ructb = "0".zfill(7) + "5" + province.code.zfill(2) + "43" + folio_imagen.zfill(3) + asiento_ficha.zfill(5)
    else:
        ructb = "0".zfill(6) + "5" + province.code.zfill(2) + "43" + folio_imagen.zfill(3) + asiento_ficha.zfill(6)
    return ructb



def calculate_vd(old_ruc: bool, ructb: str) -> str:
    j = 2
    nsuma = 0

    for c in reversed(ructb):
        if old_ruc and j == 12:
            old_ruc = False
            j -= 1

        nsuma += j * (ord(c) - ord("0"))
        j += 1
    r = nsuma % 11
    return str(11 - r) if r > 1 else str(0)


def _digitDV(sw, ructb):
    # rutina calcula dv
    j = 2
    nsuma = 0

    for c in reversed(ructb):
        if sw and j == 12:
            sw = False
            j -= 1

        nsuma += j * (ord(c) - ord("0"))
        j += 1
    r = nsuma % 11
    if r > 1:
        return 11 - r
    return 0



"""Console script for verification_digit_pty."""

import click


@click.command()
def main() -> None:
    """Main entrypoint."""
    click.echo("verification-digit-pty")
    click.echo("=" * len("verification-digit-pty"))
    click.echo("Library to calculate the verification digit from RUC and cedulas in Panama.")


if __name__ == "__main__":
    main()  # pragma: no cover



import pytest

from verification_digit_pty.calculators.natural import calculate_verification_digit


class TestCalculateVerificationDigit:
    def testt_errors(self) -> None:
        assert calculate_verification_digit("") == ""
        assert calculate_verification_digit("E") == ""

    @pytest.mark.parametrize(
        "input_value, expected_output",
        [
            ("61302-14-123411", "22"),
            ("1102-85-117211", "95"),
            ("2486589-1-816994", "62"),
            ("1830234-1-710357", "82"),
            ("65219-68-360495", "20"),
            ("41369-85-283456", "73"),
            ("11947-1027-0229562", "71"),
            ("11947-1-0229562", "42"),
            ("64296-75-357434", "00"),
            ("203141-1-17214", "60"),
            ("1075137-1-553125", "18"),
        ],
    )
    def test_juridical(self, input_value, expected_output) -> None:
        assert calculate_verification_digit(input_value) == expected_output

    @pytest.mark.parametrize(
        "input_value, expected_output",
        [
            ("8-442-445", "08"),
            ("PE-10-442", "50"),
            ("N-45-832", "58"),
            ("E-12-342", "10"),
            ("1AV-432-658", "31"),  # TODO Changed from 96 to 31 with the new implementation
            ("4PI-234-123", "31"),  # TODO Changed from 96 to 31 with the new implementation
        ],
    )
    def test_natural(self, input_value, expected_output) -> None:
        assert calculate_verification_digit(input_value) == expected_output

    @pytest.mark.parametrize(
        "input_value, expected_output",
        [
            ["155720753-2-2022", "39"],
            ["2588017-1-831938", "20"],
            ["1489806-1-645353", "68"],
            ["1956569-1-732877", "00"],
            ["797609-1-493865", "12"],
            ["15565624-2-2017", "63"],
        ],
    )
    def test_juridical_old(self, input_value, expected_output) -> None:
        """Tests for `verification_digit_pty` package from Panama-RUC-DV-Calculator."""
        # TODO is not raising error.    ["0-0-0", "19"],  # Error
        assert calculate_verification_digit(input_value) == expected_output

    # @pytest.mark.skip(reason="Not supported yet")
    @pytest.mark.parametrize(
        "input_value, expected_output",
        [
            ["0-NT-0-0", "31"],
            ["8-NT-1-13656", "43"],
            ["1-NT-45-56544", "03"],
            ["5-NT-478-2351", "94"],
            ["7-NT-102-33575", "03"],
            ["11-NT-958-2182101", "82"],
            ["8-NT-1-1234567", "49"],
            ["11-NT-958-218210", "73"],
            ["11-NT-958-2182104", "82"],
            ["8-NT-1-123456", "52"],
        ],
    )
    def test_juridical_nt(self, input_value, expected_output) -> None:
        """Tests for `verification_digit_pty` package from Panama-RUC-DV-Calculator."""
        # TODO this is not working.
        assert calculate_verification_digit(input_value) == expected_output

    @pytest.mark.parametrize(
        "input_value, expected_output",
        [
            ["E-0-0", "75"],
            ["E-8-127702", "16"],
            ["E-8-127703", "05"],
            ["E-8-127702", "16"],
            ["E-8-12770", "72"],
            ["E-1234-12770", "98"],
            ["E-1235-12770", "23"],
            ["E-1-11", "63"],
            ["E-7824-53189", "90"],
            ["E-9624-41065", "80"],
            ["E-6521-53249", "99"],
            ["E-5056-27219", "16"],
            ["E-123-1277012", "65"],
            ["E-8-96407", "29"],
            ["E-1234-123456789", "26"],
        ],
    )
    def test_natural_e(self, input_value, expected_output) -> None:
        """Tests for `verification_digit_pty` package from Panama-RUC-DV-Calculator."""
        assert calculate_verification_digit(input_value) == expected_output

    @pytest.mark.parametrize(
        "input_value, expected_output",
        [
            ["0AV-0-0", "10"],
            ["8AV-1-196", "90"],
            ["2AV-1234-12345", "26"],
            ["2AV-1234-123", "33"],
            ["2AV-123-123456", "28"],
            ["8AV-123-123456", "78"],
            ["2AV-1234-1234", "02"],
        ],
    )
    def test_natural_av(self, input_value, expected_output) -> None:
        """Tests for `verification_digit_pty` package from Panama-RUC-DV-Calculator."""
        assert calculate_verification_digit(input_value) == expected_output

    @pytest.mark.parametrize(
        "input_value, expected_output",
        [
            ["0PI-0-0", "57"],
            ["13PI-1-196", "58"],
            ["8PI-1-80", "05"],
            ["8PI-23-65", "91"],
            ["2PI-23-65", "41"],
            ["2PI-123-1234", "41"],
            ["2PI-1234-12345", "26"],
            ["2PI-1234-123", "33"],
            ["2PI-123-123456", "65"],
            ["2PI-1234-1234", "02"],
            ["8PI-1234-1234", "02"],
            ["8PI-1234-12345", "26"],
        ],
    )
    def test_natural_pi(self, input_value, expected_output) -> None:
        """Tests for `verification_digit_pty` package from Panama-RUC-DV-Calculator."""
        assert calculate_verification_digit(input_value) == expected_output

    @pytest.mark.parametrize(
        "input_value, expected_output",
        [
            ["N-0-0", "76"],
            ["N-19-1821", "11"],
            ["N-1-24", "89"],
            ["N-1234-12345", "00"],
            ["N-7824-53189", "73"],
            ["N-9624-41065", "63"],
            ["N-6521-53249", "72"],
        ],
    )
    def test_natural_n(self, input_value, expected_output) -> None:
        """Tests for `verification_digit_pty` package from Panama-RUC-DV-Calculator."""
        assert calculate_verification_digit(input_value) == expected_output

    @pytest.mark.parametrize(
        "input_value, expected_output",
        [
            ["PE-0-0", "14"],
            ["PE-1-19", "60"],
            ["PE-123-12345", "42"],
            ["PE-842-3681", "51"],
            ["PE-712-5789", "82"],
            ["PE-523-8262", "37"],
        ],
    )
    def test_natural_pe(self, input_value, expected_output) -> None:
        """Tests for `verification_digit_pty` package from Panama-RUC-DV-Calculator."""
        assert calculate_verification_digit(input_value) == expected_output



import re
from enum import Enum


class NaturalRUCLetter(Enum):
    NO_LETTER = ("", "00", "00", "No letter")
    E = ("E", "5", "66", "Foreigner")
    PE = ("PE", "75", "82", "Panamanian Foreigner ")
    N = ("N", "4", "92", "Naturalized")
    AV = ("AV", "15", "9595", "Before the system")
    PI = ("PI", "79", "9595", "Indigenous People")

    def __init__(self, letter, code, validation_code, name) -> None:
        self.letter = letter
        self.code = code
        self.validation_code = validation_code
        self.code_name = name

    @classmethod
    def from_code(cls, letter: str):
        for member in cls:
            if member.letter == letter.upper():
                return member
        raise ValueError(f"Invalid RUC letter: {letter}")

    @classmethod
    def from_part(cls, part):
        regex = re.compile(r"(\d+)?(AV|PI)")
        match = regex.match(part)
        if match:
            return cls.from_code(match.group(2))
        raise ValueError(f"Invalid RUC part: {part}.")


class Province(Enum):
    EN_BLANCO = ("", "En Blanco")
    NO_ASIGNADA = ("00", "No Asignada")
    BOCAS_DEL_TORO = ("01", "Bocas Del Toro")
    COCLE = ("02", "Cocle")
    COLON = ("03", "Colon")
    CHIRIQUI = ("04", "Chiriqui")
    DARIEN = ("05", "Darien")
    HERRERA = ("06", "Herrera")
    LOS_SANTOS = ("07", "Los Santos")
    PANAMA = ("08", "Panama")
    VERAGUAS = ("09", "Veraguas")
    GUNA_YALA = ("10", "Guna Yala, Madugandí y Wargandí")
    EMBER_WOUNAAN = ("11", "Embera Wounaan")
    NGABE_BUGLE = ("12", "Ngabe Bugle")
    PANAMA_OESTE = ("13", "Panama Oeste")

    def __init__(self, code, description) -> None:
        self.code = code
        self.province_name = description

    @classmethod
    def from_code(cls, code):
        for member in cls:
            if member.code == code.zfill(2):
                return member
        raise ValueError(f"Invalid province code: {code}")

    @classmethod
    def from_part(cls, part: str):
        regex = re.compile(r"(\d+)?(AV|PI)")
        match = regex.match(part)
        if match:
            return cls.from_code(match.group(1))
        raise ValueError(f"Invalid province part: {part}")
